from odoo import models, fields, api, http, Command, _
from odoo.addons.saas_worker.auth_util import SaasAuth
from odoo.tools import float_utils, DEFAULT_SERVER_DATETIME_FORMAT, mail, cloc
from odoo.tools.misc import file_open
from odoo.tools.mimetypes import guess_mimetype
from odoo.exceptions import AccessDenied, ValidationError
from odoo.modules.db import create_categories
from odoo.modules.module import get_manifest, get_module_icon, README
from odoo.addons.loempia import util
from odoo.addons.saas_worker.util import hash62, humanize_version
from odoo.addons.saas_worker.const import USER, BUYAPPS_HOST, APPS_KEY, APPS_URL, TIMEOUT, LOCAL


from odoo.addons.base.models.ir_module import Module as ir_module_module
LICENSES = ir_module_module.license.args['selection'][:]


import os
import io
import hmac
import base64
import shutil
import logging
import platform
import requests
import tempfile
import threading
import lxml.html
import subprocess
from math import ceil
from PIL import Image
from hashlib import sha1,sha256
from contextlib import closing
from urllib.parse import urlparse
from datetime import date, datetime
from dateutil.relativedelta import relativedelta


_logger = logging.getLogger(__name__)

SUPPORTED_CURRENCIES = ['EUR', 'USD']

# link sanitization in descriptions
ALLOWED_EXT = ('.pdf',)
ALLOWED_PATH = ('/apps',)
ALLOWED_ENDPOINTS = ('www.odoo.com', 'apps.odoo.com', 'youtube.com', 'youtu.be')
ALLOWED_PROTOCOLS = ('mailto:', 'skype:')
ALLOWED_MIMETYPES = ('image/png', 'image/gif', 'image/jpeg')

class ModuleTemplate(models.Model):
    _name = 'loempia.module.template'
    _description = 'Apps Template'
    _inherit = ['mail.thread']

    name = fields.Char('Module Technical Name', required=True, index=True, readonly=True, tracking=True)
    maintainer_id = fields.Many2one('res.users', string='Maintainer', required=True, index=True, tracking=True)
    display_name = fields.Char('Module Display Name', compute='_compute_display_name')

    module_ids = fields.One2many("loempia.module", "template_id")
    commercial_version_ids = fields.One2many("loempia.module", compute="_compute_commercial_version_ids")

    download_ids = fields.One2many('loempia.module.download', 'template_id')
    download_count_all = fields.Integer('Template downloads count', compute='_compute_dl_all', store=True)
    download_count_all_recent = fields.Integer('Recent template downloads count', default=0, readonly=True, index=True)

    latest_module_id = fields.Many2one('loempia.module', store=True, compute="_get_latest_module_id")
    latest_version_before_series13 = fields.Many2one('loempia.module', store=True, compute="_compute_latest_version_before_series13")

    purchase_ids = fields.One2many('loempia.module.purchase', 'template_id')
    purchases_count = fields.Integer(compute='_compute_purchases_stats', string='Purchases Count', store=True)
    purchases_count_recent = fields.Integer(string='Recent Purchases Count', default=0, readonly=True, index=True)
    sales = fields.Float(compute='_compute_purchases_stats', string='Sales Value', store=True)
    sales_recent = fields.Float(string='Recent Sales Value', default=0.0, readonly=True)

    is_blocked = fields.Boolean('Is blocked', help='Set to True to prevent people from buying the module.',
                                default=False, readonly=True, tracking=True)

    rating_ids = fields.One2many('loempia.module.rating', 'template_id')
    average_rating = fields.Integer(string="Average Rating",
                                    help='Counts the average rating when there\'s a sufficient amount (5)',
                                    compute="_compute_rating_vals", store=True, index=True)
    rating_count = fields.Integer(string="Ratings Count",
                                  help="Counts all non-zero ratings when average_rating is above 0",
                                  compute="_compute_rating_vals", store=True)

    _sql_constraints = [
        ('uniq_name_maintainer', 'UNIQUE(name,maintainer_id)', 'The name/maintainer combienation must be unique'),
    ]

    @api.depends('name', 'maintainer_id')
    def _compute_display_name(self):
        for template in self:
            template.display_name = "%s (%s)" % (template.name, template.maintainer_id.name)

    @api.depends('download_ids', 'download_ids.quantity')
    def _compute_dl_all(self):
        downloads = self.env['loempia.module.download'].read_group(
            domain=[('template_id', 'in', self.ids)],
            fields=["quantity:sum(quantity)"],
            groupby=['template_id']
        )
        data = {data['template_id'][0]: data['quantity'] for data in downloads}
        for template in self:
            template.download_count_all = data.get(template.id, 0)

    @api.model
    def _update_dl_recent_cron(self):
        time_threshold = datetime.today() - relativedelta(days=util.RECENT_THRESHOLD)
        templates = self.with_context(prefetch_fields=False).search([('download_count_all', '>', 0)])
        downloads = self.env['loempia.module.download'].read_group(
            domain=[('template_id', 'in', templates.ids), ('date', '>', time_threshold)],
            fields=["quantity:sum(quantity)"],
            groupby=['template_id']
        )
        data = {data['template_id'][0]: data['quantity'] for data in downloads}
        for template in templates:
            template.download_count_all_recent = data.get(template.id, 0)

    @api.depends('purchase_ids', 'purchase_ids.quantity', 'purchase_ids.state')
    def _compute_purchases_stats(self):
        domain = [('template_id', 'in', self.ids), ('state', 'in', ['confirmed', 'done', 'redeemed'])]
        purchases = self.env['loempia.module.purchase'].read_group(
            domain=domain,
            fields=["count:sum(quantity)", "sales:sum(price)"],
            groupby=['template_id']
        )
        data = {data['template_id'][0]: {"sales": data["sales"], "count": data["count"]} for data in purchases}
        for template in self:
            template_data = data.get(template.id, False)
            if template_data:
                template.sales = template_data.get('sales', 0)
                template.purchases_count = template_data.get('count', 0)
            else:
                template.sales = 0
                template.sales_recent = 0
                template.purchases_count = 0
                template.purchases_count_recent = 0

    @api.model
    def _update_sales_recent_cron(self):
        time_threshold = datetime.today() - relativedelta(days=util.RECENT_THRESHOLD)
        templates = self.with_context(prefetch_fields=False).search([('purchases_count', '>', 0)])
        domain = [('template_id', 'in', templates.ids), ('state', 'in', ['confirmed', 'done', 'redeemed']), ('date_order', '>', time_threshold)]
        purchases = self.env['loempia.module.purchase'].read_group(
            domain=domain,
            fields=["count_recent:sum(quantity)", "sales_recent:sum(price)"],
            groupby=['template_id']
        )
        data = {data['template_id'][0]: {"sales_recent": data["sales_recent"], "count_recent": data["count_recent"]} for data in purchases}
        for template in templates:
            template_data = data.get(template.id, False)
            if template_data:
                template.write({"sales_recent": template_data.get("sales_recent", 0), "purchases_count_recent": template_data.get("count_recent", 0)})
            else:
                template.write({"sales_recent": 0, "purchases_count_recent": 0})

    @api.depends('rating_ids.rating')
    def _compute_rating_vals(self):
        ratings = self.env['loempia.module.rating'].read_group(
            domain=[('template_id', 'in', self.ids), ('rating', '!=', 0)],
            fields=["avg_rates:avg(rating)", "rating_count:count(id)"],
            groupby=["template_id"]
        )
        default_dict = {'average_rating': 0, 'rating_count': 0}
        data = {
            data['template_id'][0]: {'average_rating': data['avg_rates'], 'rating_count': data['rating_count']}
            if data['rating_count'] >= 5 else default_dict
            for data in ratings
        }
        for template in self:
            template_data = data.get(template.id, {})
            template.write({
                'average_rating': round(template_data.get('average_rating', 0)),
                'rating_count': template_data.get('rating_count', 0),
            })

    @api.depends('module_ids.is_published', 'module_ids.is_hidden', 'module_ids.status', 'module_ids.is_deleted')
    def _get_latest_module_id(self):
        for template in self:
            valid_series_ids = template.module_ids.filtered(lambda m: not m.is_hidden and m.status in ["approved", "invalid"]
                                                                      and m.is_published and not m.is_deleted)
            if not valid_series_ids:
                valid_series_ids = template.module_ids
            template.latest_module_id = valid_series_ids and min(valid_series_ids, key=lambda m: m.series_id.sequence)

    @api.depends('module_ids.is_hidden', 'module_ids.status', 'module_ids.status', 'module_ids.is_deleted')
    def _compute_latest_version_before_series13(self):
        SEQUENCE12 = self.env.ref("loempia.series_12").sequence
        for template in self:
            all_before_13 = template.module_ids.filtered(lambda x: x.series_id.sequence >= SEQUENCE12)
            valid_before_13 = all_before_13.filtered(lambda m: not m.is_hidden and m.status in ["approved", "invalid"]
                                                               and m.is_published and not m.is_deleted)
            if not valid_before_13:
                valid_before_13 = all_before_13
            template.latest_version_before_series13 = valid_before_13 and min(valid_before_13, key=lambda m: m.series_id.sequence)

    @api.depends('module_ids', 'latest_version_before_series13')
    def _compute_commercial_version_ids(self):
        SEQUENCE12 = self.env.ref("loempia.series_12").sequence
        for template in self:
            all_after_13 = template.module_ids.filtered(lambda x: x.series_id.sequence < SEQUENCE12)
            template.commercial_version_ids = all_after_13 + template.latest_version_before_series13

    def _update_previous_version(self, module, update_product=True, update_pricing=True, update_purchases=True):
        self.ensure_one()
        SEQUENCE12 = self.env.ref("loempia.series_12").sequence
        ref_module = self.latest_version_before_series13
        modules_to_update = self.module_ids - self.commercial_version_ids
        if ref_module and modules_to_update and module.series_id.sequence >= SEQUENCE12:
            if update_product:
                self._assign_product_id(ref_module, modules_to_update)
            if update_pricing:
                self._update_old_pricing(ref_module, modules_to_update)
            if update_purchases and ref_module == module:
                self._reassign_old_purchases(ref_module, modules_to_update)

    def _assign_product_id(self, ref_module, modules_to_update):
        self.ensure_one()
        if not ref_module.product_id:
            product_id = [m.product_id for m in modules_to_update if m.product_id]
            if product_id:
                ref_module.product_id = product_id[0]
                modules_to_update.write({'product_id': ref_module.product_id})
        else:
            modules_to_update.write({'product_id': ref_module.product_id})

    def _update_old_pricing(self, ref_module, modules_to_update):
        self.ensure_one()
        for m in modules_to_update:
            retroactive = {}
            if m.currency_id != ref_module.currency_id:
                retroactive.update(currency_id=ref_module.currency_id)
            if float_utils.float_compare(m.price_single, ref_module.price_single, precision_digits=2):
                retroactive.update(price_single=ref_module.price_single)
            if retroactive:
                m.write(retroactive)

    def _reassign_old_purchases(self, ref_module, modules_to_update):
        self.ensure_one()
        old_purchases = self.env['loempia.module.purchase'].search([
            '|',
            ('module_id', 'in', modules_to_update.ids),
            ('referrer_module_id', 'in', modules_to_update.ids)
        ])
        for purchase in old_purchases:
            if purchase.module_id in modules_to_update:
                purchase.module_id = ref_module.id
            if purchase.referrer_module_id in modules_to_update:
                purchase.referrer_module_id = ref_module.id

    def toggle_block(self):
        self.ensure_one()
        self.is_blocked = not self.is_blocked
        if self.is_blocked:
            self.module_ids.status = 'blocked'
        else:
            if self.maintainer_id.banned:
                self.module_ids.status = 'banned'
            else:
                self.module_ids.status = 'invalid'
                self.module_ids.mapped("repo_id").to_scan = True

    def get_packaged_version(self, module):
        """
        This function returns all version available with the commercial version of `module`.
        For instance buying a version pre v13.0 give you access to all the previous version
        of you module.

        :param record module: current module investigated

        :returns: A recordset of all available versions of `module`
        """
        self.ensure_one()
        SEQUENCE12 = self.env.ref("loempia.series_12").sequence
        if module.series_id.sequence >= SEQUENCE12:
            modules = self.module_ids.filtered(lambda x: x.series_id.sequence >= SEQUENCE12)
        else:
            modules = module
        modules = modules.filtered(lambda x: x.is_published and x.status in ['approved', 'invalid'])
        return modules.sorted(key=lambda x: x.series_id.sequence, reverse=True)

class Module(models.Model):
    _name = 'loempia.module'
    _description = 'Apps Module'
    _inherit = ['mail.thread', 'website.seo.metadata']
    _order = 'repo_prio desc, shortdesc'
    _rec_name = 'shortdesc'
    _mail_post_access = 'read'

    def _default_currency_id(self):
        return self.env['res.currency'].search([('name', '=', 'EUR')])

    @api.depends('depend_ids')
    def _get_all_deps(self):
        memo = {}

        def closure(m):
            if m not in memo:
                memo[m] = m.depend_ids.union(*(closure(d) for d in m.depend_ids))
            return memo[m]

        for m in self:
            m.all_depend_ids = closure(m)

    def _get_free_deps(self):
        self.ensure_one()
        dids = set()
        stack = [self.depend_ids]
        while stack:
            m2m = stack.pop(0)
            for d in m2m:
                if not d.price_single:
                    dids.update([d.id])
                    stack.extend(d.depend_ids)
        return self.browse(dids)

    @api.depends('series_id.icon_path')
    def _has_icon(self):
        result = dict.fromkeys(self.ids, False)
        for module in self:
            if module.series_id.icon_path:
                icon_path = os.path.join(module.sudo().path, module.series_id.icon_path)
                if os.path.exists(icon_path):
                    result[module.id] = True
            module.has_icon = result.get(module.id)

    @api.depends('series_id.icon_path')
    def _get_icon(self):
        default_icon = get_module_icon('base').strip('/')
        with closing(file_open(default_icon, 'rb')) as fp:
            default = base64.b64encode(fp.read())
        result = dict.fromkeys(self.ids, {'icon_image': default})

        for module in self:
            if module.has_icon:
                icon_path = os.path.join(module.sudo().path, module.series_id.icon_path)
                with open(icon_path, 'rb') as fp:
                    icon = fp.read()
                    # Try to parse the image file to catch unsafe image types
                    try:
                        Image.open(io.BytesIO(icon))
                    except OSError:
                        pass
                    else:
                        result[module.id]['icon_image'] = base64.b64encode(icon)
            module.icon_image = result.get(module.id).get('icon_image')
            module.icon = '/web/image/loempia.module/' + str(module.id) + '/icon_image/48x48'

    def _get_zip(self, zip_deps=False):
        self.ensure_one()
        if os.path.exists(self.sudo().path):
            ignore = []
            repo_id = self.sudo().repo_id
            if repo_id.is_server and self.name == repo_id.series_id.product_name.lower():
                ignore = repo_id._server_zip_ignore_list(self.sudo().path)

            paths = [self.sudo().path]
            if zip_deps:
                for m in (m for m in self._get_free_deps() if not (m.sudo().repo_id.is_server or m.sudo().repo_id.is_odoo_enterprise)):
                    paths.append(m.sudo().path)
            zipfile = tempfile.TemporaryFile()
            for p in paths:
                zipfile = util.zip_dir(p, ignore=ignore, append_to=zipfile)
            return zipfile.read()

        return None

    def youpie_one_more_download(self, remote_addr=None, fake=False):
        # TODO See if it is still necessary when recaptcha will be calibrated
        try:
            # Guard the counter update with an ExclusiveWriteLock to make sure we will never
            # prevent a batch update of apps - the batch update will also grab ExclusiveWriteLock.
            self.env.cr.execute("SELECT id FROM %s WHERE id IN %%s FOR UPDATE NOWAIT" % (self._table,), (tuple(self.ids),))

            # NOTE download != install
            for m in self:
                Downloads = self.env['loempia.module.download']
                previously_seen = Downloads.search([('module_id', '=', m.id), ('uuid', '=', remote_addr)])
                if remote_addr:
                    if not previously_seen and not fake:
                        Downloads.create({'module_id': m.id, 'uuid': remote_addr, 'template_id': m.template_id.id})
                    else:
                        previously_seen.hits += 1
                        # HEAD requests or download by same IP above threshold is considered a bot
                        if fake or previously_seen.hits >= util.DL_HIT_THRESHOLD:
                            previously_seen.quantity = 0  # == cancel the downloads altogether
        except Exception:
            _logger.warning("Skipping download count update due to concurrent transaction, see DEBUG level")
            _logger.debug("Failed to increment download count for module ids: %r", self.ids, exc_info=True)
            self.env.cr.rollback()
        return True

    def _convert2html(self, html_str):
        self.ensure_one()
        html = lxml.html.fromstring(html_str)
        if not len(html):
            return False
        for element, attribute, link, pos in html.iterlinks():
            src = element.get('src')
            if src and '//' not in src and 'static/' not in src:
                # use frontend controller
                sha = sha1(str(self.write_date).encode('utf-8')).hexdigest()[0:7]
                url = '/apps/assets/{module.series_id.name}/{module.name}/{src}?{sha}'.format(module=self, src=src, sha=sha)
                # force the use of cdn if possible
                website = self.env['website'].get_current_website()
                element.set('src', website.get_cdn_url(url))

            # Transform "a" links into span so we keep the inside text
            # while breaking the link. We only tolerate links to other
            # apps on Odoo Apps because we use it for themes eCommerce
            # plugins. We also tolerate pdf links because they can be
            # used for documentation. Other extensions might be added
            # later in the form of a "valid_extensions" list if needed.
            # I chose to also keep the mailto links, but ideally we
            # should have a "contact the publisher" button.
            href = element.get('href')
            parsed = href and urlparse(href)
            if href \
                and (parsed.netloc and not parsed.netloc.endswith(ALLOWED_ENDPOINTS)
                    or not parsed.netloc and parsed.path and not parsed.path.startswith(ALLOWED_PATH))\
                and not href.endswith(ALLOWED_EXT)\
                and not href.startswith(ALLOWED_PROTOCOLS):
                element.tag = 'span'

        # avoid crappy static windows & modals
        for element in html.cssselect('[class*="modal"],[class*="position-fixed"]'):
            element.set('class', 'd-none')
            element.set('style', 'display:none')

        try:
            return mail.html_sanitize(lxml.html.tostring(html), sanitize_style=True, silent=False)
        except Exception:
            return False

    @api.depends('name', 'series_id.name', 'description')
    def _get_desc(self):
        for module in self:
            desc_file = os.path.join(module.path, 'static', 'description', 'index.html')
            #  TODO: refaire ce truc pour utiliser lxml parse et éviter les foirages d'encodage
            if os.path.isfile(desc_file):
                with open(desc_file, 'rb') as fp:
                    string_desc = fp.read()
            else:
                string_desc = util.rst2html(module.description)

            module.description_html = module._convert2html(string_desc)

    @api.depends('path')
    def _get_doc(self):
        for module in self:
            doc_file = os.path.join(module.sudo().path, 'doc', 'index.rst')
            if os.path.isfile(doc_file):
                with open(doc_file) as fp:
                    string_doc = util.rst2html(fp.read())
                module.documentation_html = module._convert2html(string_doc)
            else:
                module.documentation_html = False

    def _find_license_file(self):
        self.ensure_one()
        # Try to find the license file in the module
        license_file = os.path.join(self.sudo().path, 'LICENSE')
        if not os.path.isfile(license_file):
            # Fallback on repo level license file
            license_file = os.path.join(self.repo_id.sudo().path, 'LICENSE')
            if not os.path.isfile(license_file):
                return False
            else:
                return license_file
        else:
            return license_file

    @api.depends('path')
    def _get_license(self):
        self.ensure_one()
        license_file = self._find_license_file()
        # There are two cases where we will display our copy of OPL-1 instead
        # of the actual module license:
        # 1. The module license is OPL-1, to prevent module maintainers to
        #    display modified versions of the license.
        # 2. If no license file can be found at all, the license is other
        #    proprietary, and the version is 8.0 or older. This is a very
        #    special case where we allow people to sell proprietary modules
        #    on Odoo 8.0 and below, but these versions of Odoo won't accept
        #    'OPL-1' as a license value in the module itself.
        opl_legacy_fallback = (
            self.license == 'Other proprietary'
            and self.series_id.sequence >= self.env.ref('loempia.series_8').sequence
            and not (license_file and os.path.isfile(license_file))
        )
        if self.license == 'OPL-1' or opl_legacy_fallback:
            loempia_path = get_manifest('loempia')['addons_path']
            license_file = os.path.join(loempia_path, 'loempia/static/examples/LICENSE')
        # Despite all the above, it's still possible that no license file can be
        # displayed for this module. For example: 'Other proprietary' for Odoo 9
        # with no license file in the module, there is no way to resolve this.
        if not license_file:
            self.license_file = False
            return
        try:
            with open(license_file) as fp:
                self.license_file = fp.read()
        except Exception:
            self.license_file = False

    @api.depends('license', 'license_file')
    def _get_license_display(self):
        for module in self:
            module.license_display = module.license in ['OPL-1', 'OEEL-1', 'Other OSI approved licence', 'Other proprietary'] and module.license_file

    @api.depends('changelog')
    def _get_changelog_html(self):
        for module in self:
            if module.changelog:
                module.changelog_html = util.rst2html(module.changelog)
            else:
                module.changelog_html = ''

    @api.depends('name', 'repo_id.snap_path')
    def _path(self):
        for module in self:
            module.path = os.path.join(module.repo_id.snap_path, module.name)

    def _get_download_code(self, validity_date=None, product_id=None):
        """Generate a temporary download code for a module.
        :param validity_date: date up to the code will be valid (default: 7 days)
        :param product_id: product id (integer) used to purchase the module
        :return: download code
        :rval: str
        """
        self.ensure_one()
        if validity_date is None:
            validity_date = date.today() + relativedelta(days=7)
        if product_id is None:
            product_id = self.product_id
        message = 'apps-download-token-%s-%s-%s' % (validity_date, self.name, product_id)
        return hmac.new(APPS_KEY.encode(), message.encode(), sha256).hexdigest()

    def _verify_download_code(self, dl_code, validity_date=None, product_id=None):
        self.ensure_one()
        computed_dl_code = self._get_download_code(validity_date=validity_date, product_id=product_id)
        return hmac.compare_digest(dl_code, computed_dl_code)

    @api.depends('name', 'version')
    def _temp_dl_url(self):
        host = self.env['ir.config_parameter'].get_param('web.base.url', default='')
        validity_date = date.today() + relativedelta(days=7)
        for module in self:
            dl_code = module._get_download_code(validity_date=validity_date)
            url = '{host}/loempia/download/tmp/{module.name}/%s/%s/%s.zip' % (validity_date, module.product_id, dl_code)
            module.temp_download_url = url.format(host=host, module=module)

    @api.depends('price_single', 'depend_ids.price')
    def _get_bundled_price(self):
        for module in self:
            module.price = module.price_single + sum(m.price_single for m in module.all_depend_ids if m.price_single)

    @api.depends('price')
    def _compute_displayed_price(self):
        pricelist_id = self.env['loempia.pricelist'].browse(self._context.get('pricelist_id')) or self.env['loempia.pricelist'].get_default_pricelist()
        for module in self:
            module.displayed_price = module.currency_id._convert(module.price, pricelist_id.currency_id, self.env.user.company_id, fields.Date.today())

    @api.depends('lines_of_code', 'depend_ids.lines_of_code')
    def _get_bundled_lines(self):
        for module in self:
            module.total_lines_of_code = module.lines_of_code + sum(m.lines_of_code for m in module.all_depend_ids if m.lines_of_code)

    @api.depends('total_lines_of_code')
    def _compute_maintenance_cost(self):
        pricelist_id = self.env['loempia.pricelist'].browse(self._context.get('pricelist_id')) or self.env['loempia.pricelist'].get_default_pricelist()
        default_batch_size = self.env['ir.config_parameter'].sudo().get_param('loempia.lines_batch_size', '100')
        for module in self:
            number_of_batches = ceil(module.total_lines_of_code / float(default_batch_size))
            module.maintenance_cost = pricelist_id.price_per_batch * number_of_batches

    @api.depends('icon_image', 'image_ids', 'description', 'description_html', 'license', 'license_file', 'ratings_ids', 'template_id.average_rating', 'status')
    def _compute_quality_score(self):
        for module in self:
            score = 5
            score_details = []

            if not module.has_icon:
                score -= 1
                score_details.append('This module has no icon.')

            if not module.image_ids:
                score -= 1
                score_details.append('This module has no cover image (thumbnail).')

            if module.license == 'Other proprietary' and not module._find_license_file():
                score -= 1
                score_details.append('This module is missing a license key in the manifest and/or a LICENSE file.')

            if module.template_id.average_rating and module.template_id.average_rating < 3:
                score -= 1
                score_details.append('This module has a low average rating.')

            # We prefer html descriptions because they are supposed to look better
            desc_file = os.path.join(module.sudo().path, 'static', 'description', 'index.html')
            if not os.path.isfile(desc_file):
                score -= 1
                score_details.append('This module has a non-html description.')

            if module.status == 'invalid':
                score -= 1
                score_details.append("This module scan triggered some errors. Please consult the repository dashboard for more details.")

            module.score = max(score, 0)
            module.score_details = '\n'.join(score_details)

    def _get_commercial_series_module_id(self):
        """
        This method will return the right version of the module to use for sales purpose.
        In particular, we now apply the following rule: All sales prior to 13.0 gives you
        access to all versions prior to 13.0. Afterwards, we sell each version independently.
        """
        self.ensure_one()
        if self.series_id.sequence >= self.env.ref("loempia.series_12").sequence:
            return self.template_id.latest_version_before_series13
        return self

    @api.depends('name', 'series_id.name')
    def _generate_hash(self):
        for module in self:
            module.download_hash = hash62(module.series_id.name + '#' + module.name)

    @api.depends('download_ids', 'download_ids.quantity')
    def _get_dl_count(self):
        for module in self:
            module.download_count = sum(module.download_ids.mapped('quantity'))

    @api.depends('repo_id', 'repo_id.url', 'depend_ids')
    def _compute_edition(self):
        enterprise_repos = ('ssh://git@github.com/odoo/enterprise.git#',) + {
            'chs': ('/Users/chs/devel/odoo/enterprise/master#',),
            'Malorne': ('/Users/Malorne/Odoo/enterprise#',),
            'mge': ('/home/mge-odoo/Dev/enterprise#',),
        }.get(USER, ())
        for module in self:
            if module.repo_id.url.startswith(enterprise_repos) or any(module.depend_ids.mapped(lambda m: m.edition == 'enterprise')):
                module.edition = 'enterprise'
            else:
                module.edition = 'community'

    @api.depends('all_depend_ids')
    def _compute_is_iap(self):
        for module in self:
            module.is_iap = 'iap' in module.all_depend_ids.mapped('name')

    @api.depends('is_theme', 'series_id.create_date', 'name')
    def _compute_page_url(self):
        for module in self:
            module_type = 'modules'
            if module.is_theme:
                module_type = 'themes'
            elif module.category_id.name in util.CATEGORIES_INDUSTRY:
                module_type = 'industries'
            module_type = 'themes' if module.is_theme else 'modules'
            module.page_url = '/apps/{type}/{module.series_id.name}/{module.name}'.format(type=module_type, module=module)

    @api.depends('name', 'series_id')
    def _compute_display_name(self):
        for mod in self:
            mod.display_name = "%s (%s)" % (mod.name, mod.series_id.name)

    name = fields.Char('Module Name', required=True, index=True, readonly=True, tracking=True)
    display_name = fields.Char('Module Display Name', compute='_compute_display_name')
    page_url = fields.Char('Mdule page url', compute='_compute_page_url')
    version = fields.Char('Version', required=True, tracking=True)
    category_id = fields.Many2one('ir.module.category', 'Category', readonly=True, index=True, tracking=True)   # reuse existing model
    repo_id = fields.Many2one('loempia.repo', string='Repository', required=True, tracking=True, ondelete='cascade')
    series_id = fields.Many2one('loempia.series', related='repo_id.series_id', string='Series', readonly=True, store=True)
    repo_maintainer_id = fields.Many2one('res.users', related='repo_id.maintainer_id', readonly=True, store=True, index=True)
    repo_prio = fields.Integer(related='repo_id.priority', string='Repository Priority', readonly=True, store=True)
    #'vcs_revision_id': fields.char('Revision ID', groups='loempia.group_manager')
    application = fields.Boolean('Application', readonly=True)
    auto_install = fields.Boolean('Automatic Installation', readonly=True)

    shortdesc = fields.Char('Name', readonly=False, translate=False, tracking=True, index=True)
    summary = fields.Char('Summary', readonly=True, translate=False)
    description = fields.Text("Description", readonly=True, translate=False)
    description_html = fields.Html(compute='_get_desc', string='Description HTML', sanitize_attributes=False)
    documentation_html = fields.Html(compute='_get_doc', string='Documentation HTML')
    author = fields.Char("Author", size=512, readonly=True, tracking=True)
    maintainer = fields.Char('Manifest Maintainer', size=512, readonly=True, tracking=True)
    # contributors = fields.Text('Contributors', readonly=True, tracking=True)
    website = fields.Char("Website", readonly=True, tracking=True)
    license = fields.Selection(LICENSES, 'License', readonly=True, tracking=True)
    license_file = fields.Text(compute='_get_license', string='License File')
    license_display = fields.Boolean(compute='_get_license_display', string='Display License')

    depend_ids = fields.Many2many(_name, 'loempia_module_depends', 'module_id', 'depend_id', string='Dependencies', tracking=True)   # direct dependencies
    all_depend_ids = fields.One2many(_name, compute='_get_all_deps', string='All Dependencies', readonly=True)
    release_date = fields.Date('Release Date')
    changelog = fields.Text('Changelog')
    changelog_html = fields.Html(compute='_get_changelog_html', string='Changelog HTML')
    #'security': fields.boolean('Security Update'),   # FIXME how to set it via auto_scan ?
    download_ids = fields.One2many('loempia.module.download', 'module_id', string='Downloads')
    download_count = fields.Integer('Downloads Count', compute='_get_dl_count', store=True, readonly=True)

    path = fields.Char(compute='_path', string='path')
    image_ids = fields.One2many('ir.attachment', 'res_id',
                                domain=[('res_model', '=', _name), ('mimetype', '=like', 'image/%')],
                                string='Screenshots', readonly=True)

    has_icon = fields.Boolean(compute='_has_icon', string='Module has Icon', readonly=True)
    icon = fields.Char(compute='_get_icon', string='Icon URL')
    icon_image = fields.Image(compute='_get_icon', string='Icon', readonly=True)
    icon_image_small = fields.Image("Small-sized Image", related="icon_image", max_width=64, max_height=64, help="64px * 64px")

    currency_id = fields.Many2one('res.currency', string='Currency Code (EUR or USD)', required=True, default=_default_currency_id)
    price_single = fields.Float(string='Price Single', digits=(4, 2), default=0.0, tracking=True)
    price = fields.Float(compute="_get_bundled_price", string='Price Bundled', store=True, recursive=True, index=True)
    displayed_price = fields.Float(compute="_compute_displayed_price", string="Displayed price")

    lines_of_code = fields.Integer(string='Lines of code', default=0)
    total_lines_of_code = fields.Integer(compute="_get_bundled_lines", string='Total lines of code')
    maintenance_cost = fields.Float(compute="_compute_maintenance_cost", string="Estimated maintenance cost")
    approved = fields.Boolean('Approved', tracking=True)     # filter for promoted/certified apps + auto_install triggers
    edition = fields.Selection([
        ('community', 'Community'),
        ('enterprise', 'Enterprise')
    ], string="Odoo Edition", compute='_compute_edition', store=True, help="Required Odoo Edition for this module.")

    download_hash = fields.Char(compute='_generate_hash', string='Download Hash')
    temp_download_url = fields.Char(compute='_temp_dl_url', string='Temporary Download URL', groups='loempia.group_manager')

    ratings_ids = fields.One2many('loempia.module.rating', 'module_id', string='Ratings')

    live_test_url = fields.Char("Live Test URL", readonly=True, tracking=True)

    product_id = fields.Integer('Associated Product ID', index=True, readonly=True)

    is_theme = fields.Boolean('Is a theme', help='Set to True if the module is a theme, False otherwise.', default=False)
    is_hidden = fields.Boolean('Is hidden', help='Set to True to hide the module from the website.', default=False)
    status = fields.Selection([
        ("blocked", "Blocked"),
        ("banned", "Banned"),
        ("invalid", "Invalid"),
        ("approved", "Approved")
    ], string="Status", index=True)
    is_published = fields.Boolean('Is published by the maintainer', help='Set to True to display the module on the website.', default=True, tracking=True)
    is_deleted = fields.Boolean('Deleted on remote',
        help='Set to True when the module code in not linked to any active remote branch.',
        default=False, index=True, readonly=True, tracking=True)

    purchases = fields.One2many('loempia.module.purchase', 'module_id', string="Purchases")
    support = fields.Char('Support Contract Address', required=True, readonly=True, default='apps@odoo.com')
    score = fields.Integer(compute='_compute_quality_score', store=True, string='Quality score', help="Quality score on 5.")
    score_details = fields.Char(compute='_compute_quality_score', store=True, string='Score details', help="Reasons for the score.")
    template_id = fields.Many2one('loempia.module.template', required=True, ondelete='cascade')

    is_iap = fields.Boolean(compute='_compute_is_iap', string='Requires iap services')
    is_data_module = fields.Boolean(string='Is a data module', index=True)

    @api.constrains('depend_ids')
    def _constraints_depend_ids(self):
        if not self._check_m2m_recursion('depend_ids'):
            raise ValidationError(_("Recursion found in dependencies"))

    _sql_constraints = [
        ('uniq_name_series', 'UNIQUE(name,series_id)', 'module name must be unique per series'),
    ]

    @api.model_create_multi
    def create(self, vals_list):
        for vals in vals_list:
            # Auto approve module if it comes from a repo with auto_approve enabled
            repo_id = vals.get('repo_id')
            approved = vals.get('approved')
            if repo_id and approved is not True:
                repo = self.env['loempia.repo'].browse(repo_id)
                vals['approved'] = repo.sudo().auto_approve

            template = self.env['loempia.module.template'].browse(vals.get('template_id'))
            if template.is_blocked:
                vals['status'] = 'blocked'

        return super(Module, self).create(vals_list)

    def check_mail_message_access(self, mids, operation, model_name=None):
        """ mail.message document permission rule: can post a new message if can read
            because of portal document. """
        if not model_name:
            model_obj = self
        else:
            model_obj = self.env[model_name]
        if operation == 'create':
            if self.env.user.has_group('portal.group_anonymous'):
                raise AccessDenied(_('Please log in before commenting.'))
            model_obj.check_access_rights('read')
            model_obj.browse(mids).check_access_rule('read')
        else:
            return super(Module, self).check_mail_message_access(mids, operation, model_name=model_name)

    def _toggle_publish(self):
        for module in self:
            module.is_published = not module.is_published

    def get_screenshot(self):
        self.ensure_one()
        if len(self.image_ids) > 1:
            for img in self.image_ids:
                if img.name.rsplit(".", 1)[0].endswith('_screenshot'):
                    return img
        return False

    def get_purchases_users(self):
        self.ensure_one()
        commercial_version = self._get_commercial_series_module_id()
        # If a user bought the module but its purchase is now cancelled we should still displayed the confirmed purchase tag
        # Because, the customer might have tested the module
        not_draft_purchases = self.env['loempia.module.purchase'].search([('module_id', '=', commercial_version.id), ('state', '!=', 'draft')])
        return not_draft_purchases.mapped('user_id.partner_id')

    @api.model
    def _update(self, repo, path, values, deps=set()):
        all_series = self.search([('name', '=', values.get('name'))])
        modules = all_series.filtered(lambda x: x.series_id == repo.series_id)
        module = modules and modules[0]

        if module:
            if not module._is_name_available_for_serie(repo):
                raise util.ModuleRepoError('Module %s for series %s belongs to another repository. Please select a free module name\
                    or user the right user account.' % (values.get('name'), repo.series_id.name))

            if module.approved and not repo.auto_approve:
                return module

        elif repo.maintainer_id.banned:
            raise util.ModuleRepoError('Module %s: You cannot register a new module during an active period of ban.' % values.get('name'))

        elif not repo.auto_approve:
            # check that technical name is not already used on another repo
            is_techname_available = all_series._is_techname_available(repo, values)
            if not is_techname_available:
                raise util.ModuleRepoError('Module %s already exists for another serie. Please select a free module name\
                    or use the right user account.' % values.get('name'))

        module_dep_ids = module and set(module.depend_ids.ids) or set()
        if module_dep_ids != deps:  # Avoid to redefine the dependencies unnecessarily as it triggers the recomputaion of several fields
            to_add = [(4, dep, 0) for dep in deps if dep not in module_dep_ids]
            to_remove = [(3, dep, 0) for dep in module_dep_ids if dep not in deps]
            values['depend_ids'] = to_add + to_remove

        self._validate_license(repo, values)

        deps = self.search([('name', 'in', values.pop('depends', [])), ('series_id', '=', repo.series_id.id)])
        all_deps = deps + deps.mapped('all_depend_ids')
        deps_licenses = set(dep.license for dep in all_deps if not dep.repo_id.is_server)

        if any([not li.startswith(util.LICENSES_MAP[values.get('license')]) for li in deps_licenses]):
            raise util.ManifestError('The license %s conflicts with one of its dependencies. \
                Please make sure that you comply with the licenses compatibility.\
                See our FAQ at %s/apps/faq#maintainer_faq_04 .' % (values.get('license'), APPS_URL))

        values['changelog'] = util.get_changelog(path)

        # Force banned based on the maintainer's status
        if repo.maintainer_id.banned:
            values['status'] = 'banned'

        # keep the blocked state of the latest existing version of the module.
        values['status'] = all_series\
             and min(all_series, key=lambda m: m.series_id.sequence).status == 'blocked' and 'blocked'\
             or values['status']

        images = values.pop('images', [])
        if module:
            template = module.template_id
            if module.version != values.get('version'):
                util.chatter_log(module, 'info')('update module %s to version %s', values.get('name'), values.get('version'))
            # Avoid recomputation of several fields
            if module.repo_id != repo:
                values['repo_id'] = repo.id
            if module.category_id.id == values.get("category_id", None):
                values.pop('category_id', None)
            if module.status in ['blocked', 'banned']:
                values.pop('is_hidden', None)
            if module not in template.commercial_version_ids:
                [values.pop(key, None) for key in ["icon_image", "price_single", "currency_id"]]
            elif not float_utils.float_compare(module.price_single, values.get('price_single', 0.0), precision_digits=2):
                [values.pop(key, None) for key in ['price_single', 'currency_id']]
            module.write(values)
        else:
            module_name = values.get('name')
            template = self._assign_template(module_name, repo)
            values.update({'template_id': template.id, 'repo_id': repo.id})
            module = self.create(values)

        if os.path.exists(module.path):
            if os.path.islink(module.path):
                os.remove(module.path)
            else:
                shutil.rmtree(module.path)

        if platform.system() == 'Darwin':
            subprocess.call('cd %s && find . -print0 | cpio -p -0 -d -al %s' % (path, module.path), shell=True, stderr=subprocess.PIPE)
        else:
            subprocess.call(['cp', '-al', path, module.path])

        if not repo.is_server and not repo.is_odoo_enterprise:
            if module._can_maintenance_be_charged():
                try:
                    is_industry_module = (
                        module.category_id.name in util.CATEGORIES_INDUSTRY
                        and module.repo_maintainer_id == self.env.ref('base.user_admin')
                        and module.is_data_module
                    )
                    if is_industry_module:
                        module.lines_of_code = util._count_lines(path)
                    else:
                        c = cloc.Cloc()
                        c.count_path(module.path)
                        module.lines_of_code = c.code.get(module.name)
                except Exception:
                    module.lines_of_code = 0
            else:
                module.lines_of_code = 0

        module._create_attachments_for_module(path, images)
        template._update_previous_version(module)
        module._compute_quality_score()  # TODO check if it's still needed
        return module

    def _can_maintenance_be_charged(self):
        self.ensure_one()
        SEQUENCE11 = self.env.ref("loempia.series_11").sequence
        return self.series_id.sequence <= SEQUENCE11

    @api.model
    def create_from_path(self, repo_id, path, modules_to_update):
        """
            create or update a module using the given path.
        """
        repo = self.env['loempia.repo'].browse(repo_id)
        path = os.path.normpath(path)
        readme_path = [os.path.join(path, x) for x in README if os.path.isfile(os.path.join(path, x))]
        values = util.get_values_from_manifest(path, readme_path, repo, SUPPORTED_CURRENCIES)
        from_currency = self.env['res.currency'].search([('name', '=', values.pop('currency', None))])
        to_currency = self.env['res.currency'].search([('name', '=', 'EUR')])
        if from_currency and to_currency:
            values.update({
                'price_single': from_currency._convert(values.get('price_single'), to_currency, self.env.user.company_id, fields.Date.today()),
                'currency_id': to_currency.id
            })

        deps = self._get_dependencies(repo, values, modules_to_update)
        category_id = create_categories(self.env.cr, values.pop('categories', None))
        values['category_id'] = category_id

        values['status'] = 'approved'  # set by default if this is not the case, it will be updated afterwards
        values['is_deleted'] = False   # Is we make it so far, the module is present in the repository
        values['is_data_module'] = self._check_data_module(path, deps)
        return self._update(repo, path, values, deps)

    @api.model
    def _check_data_module(self, path, deps):
        dependencies = self.env['loempia.module'].browse(list(deps))
        for _, _, files in os.walk(path):
            if any(f not in ('__manifest__.py', '__init__.py') and f.endswith('.py') for f in files):
                return False
            if any(not mod.is_data_module and mod.repo_id.maintainer_id.id != self.env.ref('base.user_admin').id
                   for mod in dependencies):
                return False
        return True

    @api.model
    def _assign_template(self, module_name, repo):
        template = self.env['loempia.module.template'].search([('name', '=', module_name), ('maintainer_id', '=', repo.maintainer_id.id)])
        if not template:
            template = self.env['loempia.module.template'].create({'name': module_name, 'maintainer_id': repo.maintainer_id.id})
        return template

    @api.model
    def _get_dependencies(self, repo, values, modules_to_update):
        deps = set()
        module_dependencies = [repo.series_id.product_name.lower() if dpd == 'base' else dpd for dpd in values.get('depends', [])]
        dpds = self.search_read([('name', 'in', module_dependencies), ('series_id', '=', repo.series_id.id)], ['name'])
        dpds_dict = {dpd['name']: dpd['id'] for dpd in dpds}  # With fixed serie, name can be used as identifier
        for module_dpd in module_dependencies:
            if module_dpd in modules_to_update:
                raise util.DependencyError('%s is updated before its dependency %s. Throw back to the end of the queue.' % (values.get('name'), module_dpd))
            if module_dpd not in dpds_dict:
                raise util.DependencyError('%s: unmet dependency %s for series %s. Dependency might be invalid.' % (values.get('name'), module_dpd, repo.series_id.name))
            deps.add(dpds_dict[module_dpd])
        return deps

    def _is_name_available_for_serie(self, repo):
        self.ensure_one()
        if self.repo_id != repo:
            if self.repo_maintainer_id != repo.maintainer_id or self.approved:
                return False
            else:
                _logger.info('Module %s moved from repo %s to %s.', self.name, self.repo_id.url, repo.url)
        return True

    def _is_techname_available(self, repo, values):
        # Only the last known maintainer can register the newest version of his module.
        # Any attempt to go around this should be manually treated by the support to
        # assess the legitimacy of the request.
        techname_used = self.filtered(lambda x: x.series_id != repo.series_id)
        if techname_used:
            latest_version = min(techname_used, key=lambda m: m.series_id.sequence)
            if repo.maintainer_id != latest_version.repo_maintainer_id:
                return False
        return True

    @api.model
    def _validate_license(self, repo, values):
        # override for internal modules which don't have a license key in their manifest.
        # This does not apply to the themes as they are published under OPL-1 which is the default behaviour.
        license = values.get('license')
        if repo.is_server:
            # pre 9.0 license
            if repo.series_id.sequence > self.env.ref('loempia.series_9').sequence:
                license = 'AGPL-3'
            else:
                license = 'LGPL-3'
        elif repo.is_odoo_enterprise:
            license = 'OEEL-1'

        #override for OCA's modules when they do not pass a license file in the manifest
        if repo.maintainer_id.id == 34837:  # OCA's user_id
            if values.get('license') == 'Other proprietary':
                license = 'LGPL-3'

        if not license.startswith(tuple(util.LICENSES_MAP.keys())):
            raise util.ManifestError('Invalid license: %r. See our FAQ at %s/apps/faq#maintainer_faq_04 .' % (license, APPS_URL))

        if repo.series_id.restrict_to_license:
            if license != repo.series_id.restrict_to_license:
                raise util.ManifestError('Invalid license: %r. See our FAQ at %s/apps/faq#maintainer_faq_04 .' % (license, APPS_URL))

        if license in ['OPL-1', 'OEEL-1']:
            if repo.series_id.sequence >= self.env.ref('loempia.series_8').sequence:
                raise util.ManifestError("""Invalid license for version 8.0 and below: %r.
                    To use a proprietary license on version 8.0 and below, please set the
                    license to 'Other proprietary' and we will display the OPL-1 license
                    in the appstore, unless you already provided a custom license file.
                """ % (license,))

            elif license in ['OPL-1', 'OEEL-1'] and repo.series_id.sequence >= self.env.ref('loempia.series_9').sequence and not repo.auto_approve:
                raise util.ManifestError("""Please change the license of your module from
                    OEEL-1 to 'Other proprietary'. OEEL-1 was not the right value for
                    community apps and we sincerely apologize for advising to use it instead
                    of OPL-1. However, since OPL-1 was not included in the license selection
                    field when 9.0 shipped, the only correct option for 9.0 is to use
                    'Other proprietary'. We will then display the OPL-1 license in the appstore
                    for your module, unless you already provided a custom license file. In
                    version 10.0, you will be able to correctly set the license to 'OPL-1'.
                    Please note that, in practice, it doesn't change anything to the rights
                    granted to your module because, even though we used to use OEEL-1, the
                    license file that was displayed was the OPL-1 one. What was wrong was
                    the label of the license. Your module was already in OPL-1, but it was
                    wrongly labeled as OEEL-1 due to a mistake from our part.""")

            elif license == 'OEEL-1' and not repo.auto_approve:
                raise util.ManifestError("""Please change the license of your module from OEEL-1
                to 'OPL-1'. OEEL-1 was not the right value for community apps and we sincerely
                apologize for advising to use it instead of OPL-1.""")

        values.update({'license': license})

    def _create_attachments_for_module(self, path, images):
        self.ensure_one()
        # First delete previous images
        for i in self.image_ids:
            i.unlink()
        # Create the images attachments in reverse order to have
        # image_ids attachments in same order as the manifest !
        Attach = self.env['ir.attachment']
        for image in reversed(images):
            image_path = os.path.realpath(os.path.normpath(os.path.join(path, image)))
            if not os.path.isfile(image_path) or not image_path.startswith(path):
                continue
            with open(image_path, 'rb') as fp:
                data = fp.read()
            mimetype = guess_mimetype(data)
            if mimetype not in ALLOWED_MIMETYPES:
                continue
            image_name = os.path.basename(image)
            Attach.create({
                'name': image_name,
                'datas': base64.b64encode(data),
                'res_model': self._name,
                'res_id': self.id,
            })

    @api.model
    def _create_server_module(self, repo):
        assert repo.is_server

        path = os.path.join(repo.path, repo.series_id.server_path)
        _logger.debug('create server module from path: %r on %s repo %s', path, repo.vcs, repo.url)

        path = os.path.normpath(path)
        releasepy = os.path.join(path, 'release.py')
        if not os.path.isfile(releasepy):
            raise ValueError('invalid path: %s' % releasepy)

        release = {}
        exec(open(releasepy, 'r').read(), release)

        rs = humanize_version(release.get('series', release.get('major_version')))
        if rs not in repo.series_id.name:
            raise ValueError("series missmatch: %r != %r" % (rs, repo.series_id.name))

        values = {
            'name': repo.series_id.product_name.lower(),
            'version': release.get('version'),
            'shortdesc': repo.series_id.product_name,
            'summary': '',
            'description': release.get('long_desc', ''),
            'author': release.get('author', ''),
            'website': release.get('url', ''),
            'license': release.get('license'),
            'status': 'blocked',  # Server module should not be available from the Apps store
            'is_hidden': True,  # Technical module, we don't want to show it in search results
        }

        return self._update(repo, path, values)

    @api.model
    def find_auto_install_modules(self, series_id, mods_installed_data, mods_to_install):
        """Return the directly auto_installable modules based on the modules_to_install and
           the currently installed data.

        :param int series_id: series id
        :param dict mods_installed_data: map of module names available on the instance requesting the download
            to a map of its detailed info, e.g.::

                 {'sale': {
                              'name': 'openerp',
                              'installed_version': '7.0.1.1', # the version on disk
                              'latest_version': '7.0.1.0', # the version in the database
                              'state': 'installed',
                          },
                 }

            the ``base`` module is included with the server in a special module called 'openerp'.
        :return: list of dicts with module 'name' and 'version' keys for modules to auto-install
        """
        satisfied_states = frozenset(('installed', 'to install', 'to upgrade'))
        mods_installed = [m for (m, m_data) in mods_installed_data.items()
                          if m_data['state'] in satisfied_states] or ['base']

        # auto-installed/glue modules that are approved
        candidates = self.search([('series_id', '=', series_id), ('auto_install', '=', True),
                                  ('name', 'not in', list(set(mods_installed + mods_to_install))),
                                  ('approved', '=', True)])
        auto_install_modules = []
        _logger.debug('Candidates for auto_install: %r', candidates)
        if candidates:
            # Keep those with:
            #  - all dependencies satisfied (installed or to be installed),
            #  - at least one dependency being 'to install'
            def all_depencies_satisfied(candidate):
                states = set()
                for d in candidate.all_depend_ids:
                    dname = d.name
                    if dname in mods_installed_data:
                        local_state = mods_installed_data[d.name]['state']
                        if dname in mods_to_install:
                            states.add(local_state if local_state not in ('uninstalled', 'to uninstall') else 'to install')
                        else:
                            states.add(local_state)
                    else:
                        states.add('to install' if d.name in mods_to_install else 'uninstalled')
                return states.issubset(satisfied_states) and ('to install' in states)

            auto_install_modules = filter(all_depencies_satisfied, candidates)

        _logger.debug('find_auto_install_modules results: %r', [x.name for x in auto_install_modules])
        return [{
            'id': m.id,
            'name': m.name,
            'version': m.version,
            'product_id': m.product_id,
            'price_single': m.price_single,
            'download_hash': m.download_hash
        } for m in auto_install_modules]

    def toggle_block(self):
        self.ensure_one()
        if self.status != 'blocked':
            self.status = 'blocked'
        else:
            if self.repo_maintainer_id.banned:
                self.status = 'banned'
            else:
                self.status = 'invalid'
                self.repo_id.to_scan = True

            if self.template_id.is_blocked:
                self.template_id.is_blocked = False

    def _disable_module_product(self, product_id):
        """Disactivate the product on the production and check if error is raised."""
        req = requests.post(
            BUYAPPS_HOST + '/module_product_deactivate', data={'product_id': product_id},
            auth=SaasAuth(), timeout=15,
        )
        req.raise_for_status()

    def unlink(self):
        # check user has the right to delete these modules
        self.check_access_rights('unlink')
        self.check_access_rule('unlink')

        # clear dependencies to avoid error with module inter-dependencies
        self.sudo().write({'depend_ids': [Command.clear()]})

        # Get module that we needs to disable on prod.
        product_ids_to_disactivate = self.filtered('product_id').mapped('product_id')

        # Delete carts. The above route already does it from the accounts side.
        module_ids = [m._get_commercial_series_module_id().id for m in self]
        self.env['loempia.module.purchase'].search([
            ('module_id', 'in', module_ids),
            ('state', '=', 'draft')
        ]).unlink()

        result = super(Module, self).unlink()

        for product_id in product_ids_to_disactivate:
            self._disable_module_product(product_id)

        return result

    def _default_website_meta(self):
        """
        When a link to a module is posted on social media, the preview image is set to the icon of the module if available
        """
        default_meta = super()._default_website_meta()
        if self.has_icon:
            default_meta['default_opengraph']['og:image'] = http.request.website.image_url(self, 'icon_image')
            default_meta['default_twitter']['twitter:image'] = http.request.website.image_url(self, 'icon_image', '300x300')

        return default_meta


class module_category(models.Model):
    _name = 'ir.module.category'
    _inherit = ['ir.module.category', 'website.seo.metadata']

class ModuleDepends(models.AbstractModel):
    """m2m between modules"""
    _name = 'loempia.module.depends'
    _description = 'Apps Module Dependencies'
    _table = 'loempia_module_depends'
    _log_access = False
    _auto = True

    module_id = fields.Many2one('loempia.module', string='Module', required=True, ondelete='cascade')
    depend_id = fields.Many2one('loempia.module', string='Dependency', required=True)

    _sql_constraints = [
        ('depend_id', 'CHECK(true)', "Unable to delete a module because other modules depends of it."),
    ]


class ModuleDownload(models.Model):
    _name = "loempia.module.download"
    _description = 'Apps Module Download'

    _order = 'date desc'

    @api.model
    def _default_date(self):
        return datetime.now()

    date = fields.Datetime('Download Date', required=True, readonly=True, default=_default_date, help="In the case of aggregated downloads, this is the date before which they were aggregated.")
    module_id = fields.Many2one('loempia.module', string='Module', required=True, index=True, readonly=True, ondelete="cascade")
    template_id = fields.Many2one('loempia.module.template', string='Module Template', index=True, required=True, ondelete='cascade')
    uuid = fields.Char('Hash', readonly=True, help="One time valid identifier for this download.")
    quantity = fields.Integer('Number of downloads', required=True, readonly=True, default=1)
    hits = fields.Integer('Number of hits', required=True, readonly=True, default=1)

    @api.model
    def _aggregate_downloads(self):
        time_threshold = date.today() - relativedelta(days=6*util.RECENT_THRESHOLD)  # 6 months
        # Since the aggregation erases the uuid, which serves to detect potential abuse on the download count,
        # We need to preserve the downloads that we flag as 'abuse', that is, downloads with a number of hits above DL_HIT_THRESHOLD
        domain = [('date', '<', time_threshold), ('hits', '<', util.DL_HIT_THRESHOLD)]
        # 1 fetch and process per batch
        all_dls = self.read_group(domain, ['module_id', 'template_id', 'ids:array_agg(id)', 'quantity'], groupby=['module_id', 'template_id'], lazy=False)

        batches = [all_dls[index:index + 500] for index in range(0, len(all_dls), 500)]
        for batch in batches:
            ids = []
            for module_downloads in batch:
                ids += module_downloads['ids']
                self.create({
                    'date': time_threshold,
                    'module_id': module_downloads['module_id'][0],  # read_group returns (id, name)
                    'template_id': module_downloads['template_id'][0],
                    'quantity': module_downloads['quantity'],
                    'uuid': 'aggregate',
                })
            self.browse(ids).unlink()  # Unlink the now aggregated downloads
            if not getattr(threading.current_thread(), 'testing', False):
                self.env.cr.commit()
